Include 2008 in the copyright years.
[rsync/rsync.git] / tls.c
diff --git a/tls.c b/tls.c
index 2f5b6b4..c52b25a 100644 (file)
--- a/tls.c
+++ b/tls.c
@@ -2,20 +2,21 @@
  * Trivial ls for comparing two directories after running an rsync.
  *
  * Copyright (C) 2001, 2002 Martin Pool <mbp@samba.org>
- * Copyright (C) 2003, 2004, 2005 Wayne Davison
+ * Copyright (C) 2003-2008 Wayne Davison
  *
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License version
* 2 as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 /* The problem with using the system's own ls is that some features
  * change. */
 
 #include "rsync.h"
+#include "popt.h"
+#include "lib/sysxattrs.h"
 
 #define PROGRAM "tls"
 
 /* These are to make syscall.o shut up. */
 int dry_run = 0;
+int am_root = 0;
 int read_only = 1;
 int list_only = 0;
 int preserve_perms = 0;
+int preserve_executability = 0;
+
+#ifdef SUPPORT_XATTRS
+
+#ifdef HAVE_LINUX_XATTRS
+#define XSTAT_ATTR "user.rsync.%stat"
+#else
+#define XSTAT_ATTR "rsync.%stat"
+#endif
+
+static int stat_xattr(const char *fname, STRUCT_STAT *fst)
+{
+       int mode, rdev_major, rdev_minor, uid, gid, len;
+       char buf[256];
+
+       if (am_root >= 0 || IS_DEVICE(fst->st_mode) || IS_SPECIAL(fst->st_mode))
+               return -1;
+
+       len = sys_lgetxattr(fname, XSTAT_ATTR, buf, sizeof buf - 1);
+       if (len >= (int)sizeof buf) {
+               len = -1;
+               errno = ERANGE;
+       }
+       if (len < 0) {
+               if (errno == ENOTSUP || errno == ENOATTR)
+                       return -1;
+               if (errno == EPERM && S_ISLNK(fst->st_mode)) {
+                       fst->st_uid = 0;
+                       fst->st_gid = 0;
+                       return 0;
+               }
+               fprintf(stderr, "failed to read xattr %s for %s: %s\n",
+                       XSTAT_ATTR, fname, strerror(errno));
+               return -1;
+       }
+       buf[len] = '\0';
+
+       if (sscanf(buf, "%o %d,%d %d:%d",
+                  &mode, &rdev_major, &rdev_minor, &uid, &gid) != 5) {
+               fprintf(stderr, "Corrupt %s xattr attached to %s: \"%s\"\n",
+                       XSTAT_ATTR, fname, buf);
+               exit(1);
+       }
+
+#if _S_IFLNK != 0120000
+       if ((mode & (_S_IFMT)) == 0120000)
+               mode = (mode & ~(_S_IFMT)) | _S_IFLNK;
+#endif
+       fst->st_mode = mode;
+
+       fst->st_rdev = MAKEDEV(rdev_major, rdev_minor);
+       fst->st_uid = uid;
+       fst->st_gid = gid;
+
+       return 0;
+}
+
+#endif
 
 static void failed(char const *what, char const *where)
 {
@@ -60,6 +122,10 @@ static void list_file(const char *fname)
 
        if (do_lstat(fname, &buf) < 0)
                failed("stat", fname);
+#ifdef SUPPORT_XATTRS
+       if (am_root < 0)
+               stat_xattr(fname, &buf);
+#endif
 
        /* The size of anything but a regular file is probably not
         * worth thinking about. */
@@ -74,7 +140,7 @@ static void list_file(const char *fname)
                buf.st_mode &= ~0777;
                buf.st_mtime = (time_t)0;
                buf.st_uid = buf.st_gid = 0;
-               strcpy(linkbuf, " -> ");
+               strlcpy(linkbuf, " -> ", sizeof linkbuf);
                /* const-cast required for silly UNICOS headers */
                len = readlink((char *) fname, linkbuf+4, sizeof(linkbuf) - 4);
                if (len == -1)
@@ -91,16 +157,16 @@ static void list_file(const char *fname)
        if (buf.st_mtime) {
                mt = gmtime(&buf.st_mtime);
 
-               sprintf(datebuf, "%04d-%02d-%02d %02d:%02d:%02d",
+               snprintf(datebuf, sizeof datebuf,
+                       "%04d-%02d-%02d %02d:%02d:%02d",
                        (int)mt->tm_year + 1900,
                        (int)mt->tm_mon + 1,
                        (int)mt->tm_mday,
                        (int)mt->tm_hour,
                        (int)mt->tm_min,
                        (int)mt->tm_sec);
-       } else {
-               strcpy(datebuf, "                   ");
-       }
+       } else
+               strlcpy(datebuf, "                   ", sizeof datebuf);
 
        /* TODO: Perhaps escape special characters in fname? */
 
@@ -116,18 +182,57 @@ static void list_file(const char *fname)
               datebuf, fname, linkbuf);
 }
 
+static struct poptOption long_options[] = {
+  /* longName, shortName, argInfo, argPtr, value, descrip, argDesc */
+#ifdef SUPPORT_XATTRS
+  {"fake-super",      'f', POPT_ARG_VAL,    &am_root, -1, 0, 0 },
+#endif
+  {"help",            'h', POPT_ARG_NONE,   0, 'h', 0, 0 },
+  {0,0,0,0,0,0,0}
+};
+
+static void tls_usage(int ret)
+{
+  FILE *F = ret ? stderr : stdout;
+  fprintf(F,"usage: " PROGRAM " [OPTIONS] FILE ...\n");
+  fprintf(F,"Trivial file listing program for portably checking rsync\n");
+  fprintf(F,"\nOptions:\n");
+#ifdef SUPPORT_XATTRS
+  fprintf(F," -f, --fake-super            display attributes including fake-super xattrs\n");
+#endif
+  fprintf(F," -h, --help                  show this help\n");
+  exit(ret);
+}
+
 int
 main(int argc, char *argv[])
 {
-       if (argc < 2) {
-               fprintf(stderr, "usage: " PROGRAM " DIR ...\n"
-                       "Trivial file listing program for portably checking rsync\n");
-               return 1;
+       poptContext pc;
+       const char **extra_args;
+       int opt;
+
+       pc = poptGetContext(PROGRAM, argc, (const char **)argv,
+                           long_options, 0);
+       while ((opt = poptGetNextOpt(pc)) != -1) {
+               switch (opt) {
+               case 'h':
+                       tls_usage(0);
+               default:
+                       fprintf(stderr,
+                               "%s: %s\n",
+                               poptBadOption(pc, POPT_BADOPTION_NOALIAS),
+                               poptStrerror(opt));
+                       tls_usage(1);
+               }
        }
 
-       for (argv++; *argv; argv++) {
-               list_file(*argv);
-       }
+       extra_args = poptGetArgs(pc);
+       if (!extra_args || *extra_args == NULL)
+               tls_usage(1);
+
+       for (; *extra_args; extra_args++)
+               list_file(*extra_args);
+       poptFreeContext(pc);
 
        return 0;
 }
@@ -2,20 +2,21 @@
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License version
- * 2 as published by the Free Software Foundation.
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 3 of the License, or
+ * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,